/* $Id: console.c,v 1.2 2003/09/18 20:26:35 rawsock Exp $ */

/* Intro {{{
 * ---------------------------------------------------------------- 
 *  DConnect Daemon
 * 
 *      #(@) Copyright (c) 2002, DConnect development team
 *      #(@) Homepage: http://www.dc.ds.pg.gda.pl/
 *
 *  This  program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY  or  FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You  should  have  received a copy  of  the GNU General Public
 *  License  along  with  this  program;  if  not,  write  to  the
 *  Free  Software  Foundation,  Inc.,  675  Mass  Ave,  Cambridge,
 *  MA 02139, USA. 
 *
 * ---------------------------------------------------------------- 
 * }}} */

/* include local header files {{{ */
#include "pch.h"
#include "main.h"
#include "conf.h"
#include "misc.h"
#include "net.h"
#include "console.h"
/* }}} */

struct userrec *user[MAXUSERS];	/* users table */
struct config conf;				/* configuration data */
pthread_t listen_th_cons;		/* listen thread handle for remote console */
fd_set sockfdset;				/* descriptors */
int listen_cons;				/* console listening socket */

/* addcons() - allocate and initialize new console user {{{ */
void addcons(int *sock,char *remote) {
	struct userrec *tmp;

	tmp=(struct userrec *)my_malloc(sizeof(struct userrec));
	bzero(tmp,sizeof(struct userrec));
	tmp->sock=dup(*sock);
	tmp->cons=conf.debug_level;
	if (setsockopts(&(tmp->sock),0))
		debug(DEBUG_WRNG,"setsockopts() failed.");

	/* we use select() anyway, but let's care for blocking send() */
	if (fcntl(tmp->sock,F_SETFL,O_NONBLOCK)<0)
		debug(DEBUG_WRNG,"fcntl() failed: %s.", strerror(errno));
	if (denied(&(tmp->sock))) {
		disconnect(tmp,"You're not allowed to connect. Good bye."); return; }
	if (!validhost(tmp,conf.conf_callow,remote)) {
		disconnect(tmp,NULL); return; }
	if (NUSERS>=MAXUSERS) {
		disconnect(tmp,"Users table is full! Increase MAXUSERS in main.h and recompile."); return; }

	tmp->ip=strdup(remote);
	tmp->ping=time(NULL);
	tmp->flag=1;
	tmp->ctype=strdup("Console");
	sendf(tmp,"\r\nDConnect Daemon v%s\r\n",PACKAGE_VERSION);
	sendf(tmp,"login: ");
	user[NUSERS++]=tmp;
	FD_SET(tmp->sock,&sockfdset);
} /* }}} */

/* listen_thread_cons() - listen for new connections to remote console {{{ */
void *listen_thread_cons(void *args) {
	int new_sock,remote_size;
	struct sockaddr_in local,remote;
	char *ip;

	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS,NULL);

	/* setup and bind the remote console listening socket */
	if ((listen_cons=socket(AF_INET,SOCK_STREAM,0))==-1) {
		debug(DEBUG_SYS|DEBUG_STD,"Socket error - console disabled: %s.", strerror(errno));
		pthread_cancel(listen_th_cons);
	}
	if (setsockopts(&listen_cons,1))
		debug(DEBUG_WRNG,"setsockopts() failed.");
	bzero(&local,sizeof(struct sockaddr_in));
	local.sin_family=AF_INET;
	local.sin_port=htons(conf.listen_port_cons);
	local.sin_addr.s_addr=INADDR_ANY;
	seteuid(0);
	setegid(0);
	if (bind(listen_cons,(struct sockaddr *)&local,sizeof(struct sockaddr))<0) {
		debug(DEBUG_SYS|DEBUG_STD,"Can't bind the console listening socket - console disabled: %s.", strerror(errno));
		pthread_cancel(listen_th_cons);
	}
	seteuid(conf.uid);
	setegid(conf.gid);
	if (listen(listen_cons,5)<0) {
		debug(DEBUG_SYS|DEBUG_STD,"listen() failed - console disabled: %s.", strerror(errno));
		pthread_cancel(listen_th_cons);
	}

	/* accept connections */
	while(1)
	{
		remote_size=sizeof(struct sockaddr_in);
		if ((new_sock=accept(listen_cons,(struct sockaddr *)&remote,&remote_size))<0)
		{
			if (errno!=EINTR)
				debug(DEBUG_WRNG,"accept() failed: %s.", strerror(errno));
			continue;
		}
		debug(DEBUG_NET,"Connection to console from %s:%d.",ip=inet_ntoa(remote.sin_addr),ntohs(remote.sin_port));
		addcons(&new_sock,ip);
		close(new_sock);
	}
	debug(DEBUG_WRNG,"listen_thread_cons(): this should be never reached!");
	pthread_cancel(listen_th_cons); /* never reached */
} /* }}} */

/* manage_cons() - proces console commands {{{ */
char manage_cons(struct userrec *usr,char *line) {
	const char *commands[]={
		"help",		// 0
		"conf",		// 1
		"rehash",	// 2
		"quit",		// 3
		"stat",		// 4
		"debug",	// 5
		"whois",	// 6
		"who",		// 7
		"users",	// 8
		"kick",		// 9
		"jq",		// 10
		"boost",	// 11
		"" /* indicates end of list */
	};
	const char *sep=" \t\r\n";
	char *cmd,*par;
	unsigned long long tmp;
	int count,x;

	count=strlen(line);
	strip_telnet(usr,line,&count);
	line[count]=0;

	cmd=strtok(line,sep);
	if (!cmd) {
	    if (usr->state<3) {
		sendf(usr,"\377\374\001\r\n"); deluser(usr,NULL); return 1;
	    }
	    sendf(usr,"> "); return 0;
	}
	if (!*cmd) {
	    if (usr->state<3) {
		sendf(usr,"\377\374\001\r\n"); deluser(usr,NULL); return 1;
	    }
	    sendf(usr,"> "); return 0;
	}

	/* parse login */
	if (usr->state==0) {
	    my_free(usr->nick);
	    usr->nick=strdup(cmd);
	    usr->state=1;
	    sendf(usr,"password: \377\373\001");
	    return 0;
	}

	/* parse password */
	if (usr->state==1) {
	    sendf(usr,"\377\374\001\r\n");
	    if (!validnick(usr,usr->nick)) {
		deluser(usr,NULL);
		return 1;
	    }
	    if (!validlogin(usr,usr->nick,cmd)) {
		deluser(usr,"Login incorrect.");
		return 1;
	    }
	    if (!strchr(usr->perm,'c')) {
		deluser(usr,"You have no console access permission.");
		return 1;
	    }
	    usr->state=2;
	    sendf(NULL,"$Hello %s|",usr->nick);
	    sendf(NULL,"$MyINFO $ALL %s %s$ $%s%c$%s$%Lu$|",usr->nick,usr->desc,usr->ctype,usr->flag,usr->email,usr->share);
	    sendf(usr,"\r\nLogged in. Type `help' for a command list.\r\n\r\n> ");
	    if (conf.flags & SHOW_JQ) sendf(NULL,"## %s has joined|",usr->nick);
	    usr->state=3;
	    return 0;
	}

	par=strtok(NULL,sep);
	count=0;
	while(*commands[count])
		if (!strcasecmp(cmd,commands[count])) break;
		else count++;
	if (*commands[count]) {
	    if (par) debug(DEBUG_CON,"[%s] %s %s",usr->nick,cmd,par);
	    else debug(DEBUG_CON,"[%s] %s",usr->nick,cmd);
	}
	switch(count) {
		/* 0. help {{{ */
		case 0: sendf(usr,"---------------------- help ----------------------\r\n");
			sendf(usr,"help\t\t\t- this help screen\r\r\n");
			sendf(usr,"conf\t\t\t- show current configuration\r\n");
			sendf(usr,"rehash\t\t\t- reload configuration\r\n");
			sendf(usr,"quit\t\t\t- leave the console\r\n");
			sendf(usr,"stat\t\t\t- statistics\r\n");
			sendf(usr,"debug [#]\t\t- set debug level\r\n");
			sendf(usr,"whois <nick>\t\t- show user description\r\n");
			sendf(usr,"who\t\t\t- show who is on the console\r\n");
			sendf(usr,"users\t\t\t- show all hub users\r\n");
			sendf(usr,"kick <nick> [reason]\t- disconnect a user\r\n");
			sendf(usr,"jq\t\t\t- toggle displaying of joins & quits traffic\r\n");
			sendf(usr,"boost\t\t\t- toggle info-booster\r\n");
			sendf(usr,"--------------------------------------------------\r\n");
			break; /* }}} */
		/* 1. conf {{{ */
		case 1:	sendf(usr,"------------- current configuration --------------\r\n");
			sendf(usr,"Config file: %s\r\n", conf.conf_main);
			sendf(usr,"Welcome file: %s\r\n", conf.conf_welcome);
			sendf(usr,"MOTD file: %s\r\n", conf.conf_motd);
			sendf(usr,"Bans file: %s\r\n", conf.conf_banned);
			sendf(usr,"Console users file: %s\r\n", conf.conf_cusers);
			sendf(usr,"Console allow file: %s\r\n", conf.conf_callow);
			sendf(usr,"Nicks allow file: %s\r\n", conf.conf_nallow);
			sendf(usr,"Console default debug level: %u\r\n", conf.debug_level);
			sendf(usr,"Log level: %u\r\n", conf.log_level);
			sendf(usr,"Log files dir: %s\r\n", conf.log_dir);
			sendf(usr,"Main log: %s\r\n", conf.log_main);
			sendf(usr,"Network access log: %s\r\n", conf.log_net);
			sendf(usr,"Protocol log: %s\r\n", conf.log_proto);
			sendf(usr,"Warnings log: %s\r\n", conf.log_wrng);
			sendf(usr,"Console log: %s\r\n", conf.log_con);
			sendf(usr,"Public chat log: %s\r\n", conf.log_chat);
			sendf(usr,"Main listening port: %u\r\n", conf.listen_port_main);
			sendf(usr,"Console listening port: %u\r\n", conf.listen_port_cons);
			sendf(usr,"UID: %u\r\n", conf.uid);
			sendf(usr,"GID: %u\r\n", conf.gid);
			sendf(usr,"Hub Name: %s\r\n", conf.hubname);
			sendf(usr,"User limit: %u\r\n", conf.userlimit);
			sendf(usr,"Allow broken keys: %d\r\n", conf.allow_broken_key);
			sendf(usr,"Ping timeout: %u\r\n", conf.ping_timeout);
			sendf(usr,"Registration timeout: %u\r\n", conf.register_timeout);
			sendf(usr,"Minimum data shared: %LuMB\r\n", conf.minshare);
			sendf(usr,"Minimum slots: %u\r\n", conf.minslots);
			sendf(usr,"Allow chat: %d\r\n", conf.allow_chat);
			sendf(usr,"Flags: %u\r\n", conf.flags);
			sendf(usr,"--------------------------------------------------\r\n");
			break; /* }}} */
		/* 2. rehash {{{ */
		case 2: finish(SIGHUP);
			break; /* }}} */
		/* 3. quit {{{ */
		case 3: deluser(usr,NULL);
			return 1; /* }}} */
		/* 4. stat {{{ */
		case 4: tmp=0; for(x=0;x<NUSERS;x++) tmp+=user[x]->share;
			tmp/=1024*1024; x=0; // MB
			if (tmp>=1024) {
			    tmp/=1024; x=1; // GB
			    if (tmp>=1024) { tmp/=1024; x=2; } // TB
			}
			sendf(usr,"------------------- statistics -------------------\r\n");
			sendf(usr,"Users: %d\r\n",NUSERS);
			sendf(usr,"Shared data: %Lu%cB\r\n",tmp,(x==0)?'M':((x==1)?'G':'T'));
			sendf(usr,"Uptime: %s\r\n",uptime(starttime));
			sendf(usr,"--------------------------------------------------\r\n");
			break; /* }}} */
		/* 5. debug {{{ */
		case 5: if (!par) { sendf(usr,"Current debug level: %d.\r\n",usr->cons); break; }
			x=atoi(par);
			if (!*par) goto error;
			if (x<1) x=1;
			usr->cons=x;
			break; /* }}} */
		/* 6. whois {{{ */
		case 6: if (!par) goto error;
			if (!*par) goto error;
			if ((x=nick2id(par))<0) {
			    sendf(usr,"There is no user `%s'.\r\n",par);
			    break;
			}
			sendf(usr,"---------------------- whois ---------------------\r\n");
			sendf(usr,"Nick: %s\r\n",user[x]->nick);
			sendf(usr,"IP: %s\r\n",user[x]->ip);
			sendf(usr,"Mode: %s\r\n",(user[x]->flag==1)?"online":((user[x]->flag==3)?"away":"unknown"));
			sendf(usr,"Idle: %s\r\n",uptime(user[x]->ping));
			if (user[x]->cons==0) {
			    tmp=user[x]->share;
			    tmp/=1024*1024; count=0; // MB
			    if (tmp>=1024) {
				tmp/=1024; count=1; // GB
				if (tmp>=1024) { tmp/=1024; count=2; } // TB
			    }
			    sendf(usr,"Shared: %Lu%cB\r\n",tmp,(count==0)?'M':((count==1)?'G':'T'));
			}
			sendf(usr,"E-mail: %s\r\n",user[x]->email);
			sendf(usr,"Connection: %s\r\n",user[x]->ctype);
			sendf(usr,"Description: %s\r\n",user[x]->desc);
			sendf(usr,"Permissions: %s\r\n",user[x]->perm);
			sendf(usr,"--------------------------------------------------\r\n");
			break; /* }}} */
		/* 7. who {{{ */
		case 7: sendf(usr,"USER                 FROM             IDLE\r\n");
			for (x=0;x<NUSERS;x++)
			    if ((user[x]->state==3)&&(user[x]->cons))
			    	sendf(usr,"%-20s %-15s  %s\r\n",user[x]->nick,user[x]->ip,uptime(user[x]->ping));
			break; /* }}} */
		/* 8. users {{{ */
		case 8: tmp=0;
			for (x=0;x<NUSERS;x++) {
			    if (user[x]->state>1) { sendf(usr,"%-26s",user[x]->nick); tmp++; }
			    if (tmp==3) { sendf(usr,"\r\n"); tmp=0; }
			}
			if (tmp&&(tmp<3)) { sendf(usr,"\r\n"); tmp=0; }
			break; /* }}} */
		/* 9. kick {{{ */
		case 9: if (!par) goto error;
			if (!*par) goto error;
			if ((x=nick2id(par))<0) {
			    sendf(usr,"There is no user `%s'.\r\n",par);
			    break;
			}
			if (user[x]==usr) {
			    sendf(usr,"It's simplier to use `quit'.\r\n");
			    break;
			}
			if (strchr(user[x]->perm,'a')) {
			    sendf(usr,"Administrator cannot be kicked.\r\n");
			    break;
			}
			par=strtok(NULL,"\r\n");
			if (user[x]->cons==0) {
			    sendf(user[x],"<Hub> You have been kicked off by %s",usr->nick);
			    if (par) sendf(user[x],": %s",par);
			    sendf(user[x],".|");
			}
			else {
			    sendf(user[x],"You have been kicked off by %s",usr->nick);
			    if (par) sendf(user[x],": %s",par);
			    sendf(user[x],".\r\n");
			}
			debug(DEBUG_STD,"Kicked off %s@%s by %s.",user[x]->nick,user[x]->ip,usr->nick);
			deluser(user[x],NULL);
			break; /* }}} */
		/* 10. jq {{{ */
		case 10: toggle_flag(&conf,usr,SHOW_JQ);
			 break; /* }}} */
		/* 11. boost {{{ */
		case 11: toggle_flag(&conf,usr,INFO_BOOST);
			 break; /* }}} */
		/* default {{{ */
		error:
		default: sendf(usr,"Syntax error.\r\n");
			 break; /* }}} */
	}
	sendf(usr,"> ");
	return 0;
} /* }}} */

/* VIM Settings {{{
 * Local variables:
 * tab-width: 4
 * c-basic-offset: 4
 * soft-stop-width: 4
 * c indent on
 * End:
 * vim600: sw=4 ts=4 sts=4 cindent fdm=marker
 * vim<600: sw=4 ts=4
 * }}} */

